import warnings
from typing import Union, Optional, Sequence
from weakref import WeakSet

import astropy.units as u
import glfw
import numpy as np
from astropy.units import Quantity
from numpy.typing import ArrayLike

from .context import Context
from ...simulation import PhaseToField

try:
    import OpenGL.GL as GL
    from OpenGL.GL import glViewport, glClearColor, glClear, glGenBuffers, glReadBuffer, glReadPixels, glFinish, \
        glBindBuffer, glBufferData, glBindBufferBase, glBindFramebuffer
except AttributeError:
    warnings.warn("OpenGL not found, SLM will not work")
from .patch import FrameBufferPatch, Patch, VertexArray
from ...core import PhaseSLM, Actuator, Device, Detector
from ...utilities import Transform

TimeType = Union[Quantity[u.ms], int]


class SLM(Actuator, PhaseSLM):
    """
    An OpenGL object to control a spatial light modulator connected to a graphics card.

    See :numref:`section-slms` for more information.

    """
    __slots__ = ['_vertex_array', '_frame_buffer', '_monitor_id', '_position', '_refresh_rate',
                 '_transform', '_shape', '_window', '_globals', '_frame_buffer', 'patches', 'primary_patch',
                 '_coordinate_system', '_pixel_reader', '_phase_reader', '_field_reader', '_context']

    _active_slms = WeakSet()
    """Keep track of all active SLMs. This is done for two reasons. First, to check if we are not putting two
    full-screen SLMs on the same monitor. Second, to allow sharing the OpenGL context between all SLM windows,
    so that we can use the same Patch and Texture objects on multiple SLMs simultaneously."""

    WINDOWED = 0
    patches: list[Patch]

    def __init__(self, monitor_id: int = WINDOWED, shape: Optional[tuple[int, int]] = None,
                 pos: tuple[int, int] = (0, 0), refresh_rate: Optional[Quantity[u.Hz]] = None,
                 latency: TimeType = 2, duration: TimeType = 1, coordinate_system: str = 'short',
                 transform: Optional[Transform] = None):
        """
        Constructs a new SLM window.

        Args:
            monitor_id (int): Monitor id, see :py:attr:`~monitor_id`
            shape (tuple[int,int]): Size of the window (height, width) or resolution of the fullscreen mode.
                Default value is None (recommended for full screen windows), which will use the current resolution
                of the monitor, or a standard size of 300x300 for windowed modes.
            pos (tuple[int,int]): Windowed-mode only: (y,x)-coordinate of the top-left corner of the window.
                Default value is (0,0).
            refresh_rate (Quantity[u.Hz]): Refresh rate of the SLM.
                Ignored for windowed SLMs.
                When omitted, the current refresh rate of the monitor will be used.
                Note that OpenGL does not seem to support non-integer refresh rates.
                In these cases, it is better to set the refresh rate in the OS, and not
                explicitly specify a refresh rate.
            latency (int): Time between the vertical retrace and the start of the SLM response to the new frame,
                specified in milliseconds (u.ms) or multiples of the frame period (unitless).
                see :py:attr:`~latency`
            duration (int): Time between the start of the SLM response to the newly presented frame, and the point
                where the SLM has stabilized.
                Specified in milliseconds (u.ms) or multiples of the frame period (unitless).
                see :py:attr:`~duration`
            transform (Transform): Global transformation matrix, see :py:attr:`~transform`.
                The `transform` determines how these vertex coordinates that make up the shape of a Patch (see
                :class:`Patch`) are mapped to the SLM window.
                By default, 'short' is used (see :attr:`transform`)

        Attributes:
            patches (list[Patch]): List of patches that are drawn on the SLM.
        """

        # construct window for displaying the SLM pattern
        SLM._init_glfw()
        self._assert_window_available(monitor_id)
        self._monitor_id = monitor_id
        self._position = pos
        (default_shape, default_rate, _) = SLM._current_mode(monitor_id)
        self._shape = default_shape if shape is None else shape
        self._refresh_rate = default_rate if refresh_rate is None else refresh_rate.to_value(u.Hz)
        self._frame_buffer = None
        self._window = None
        self._globals = -1
        self.patches = []
        self._context = None
        self._create_window()  # sets self._context, self._window and self._globals and self._frame_patch
        self._coordinate_system = coordinate_system
        self.transform = Transform() if transform is None else transform
        self._vertex_array = VertexArray()

        # Create a single patch for displaying phase.
        # this default patch is square 1.0, and can be accessed through the 'primary_phase_patch' attribute
        # In advanced scenarios, the geometry of this patch may be modified, or it may be replaced altogether.
        self.patches.append(Patch(self._context))
        self.primary_patch = self.patches[0]
        SLM._active_slms.add(self)

        if not isinstance(duration, Quantity):
            duration = duration * self.period
        if not isinstance(latency, Quantity):
            latency = latency * self.period

        super().__init__(duration=duration, latency=latency)
        self.update()

    def _assert_window_available(self, monitor_id) -> None:
        """
        Checks if the target monitor is available for displaying an SLM window.

        Raises:
            Exception: If a full screen SLM is already present on the target monitor.
        """
        if monitor_id == SLM.WINDOWED:
            if any([slm.monitor_id == 1 for slm in SLM._active_slms if slm is not self]):
                raise RuntimeError(
                    f"Cannot create an SLM window because a full-screen SLM is already active on monitor 1")
        else:
            # we cannot have multiple full screen windows on the same monitor. Also, we cannot have
            # a full screen window on monitor 1 if there are already windowed SLMs.
            if any([slm.monitor_id == monitor_id or
                    (monitor_id == 1 and slm.monitor_id == SLM.WINDOWED)
                    for slm in SLM._active_slms if slm is not self]):
                raise RuntimeError(f"Cannot create a full-screen SLM window on monitor {monitor_id} because a "
                                   f"window is already displayed on that monitor")
            if monitor_id > len(glfw.get_monitors()):
                raise IndexError(f"Monitor {monitor_id} not found, only {len(glfw.get_monitors())} monitor(s) "
                                 f"are connected.")

    @staticmethod
    def _current_mode(monitor_id: int):
        """Returns the current video mode resolution (height, width), refresh rate, and bit-depth of the specified
        monitor.
        For monitor_id == SLM.WINDOWED (windowed mode SLM), always returns the default window size
        of (300, 300)
        """
        if monitor_id == SLM.WINDOWED:
            monitor = glfw.get_primary_monitor()
            mode = glfw.get_video_mode(monitor)
            shape = (300, 300)
        else:
            monitor = glfw.get_monitors()[monitor_id - 1]
            mode = glfw.get_video_mode(monitor)
            shape = (mode.size[1], mode.size[0])

        return shape, mode.refresh_rate, min([mode.bits.red, mode.bits.green, mode.bits.blue])

    def _on_resize(self):
        """Updates shape and refresh rate to the actual values of the window.

        Note that these values are in pixels, which may be different from the window size because the window size is
        in screen coordinates, which may not always the same as pixels (e. g. on a retina display).

        For windowed SLMs, the refresh rate property is set to the refresh rate of the primary monitor.

        If the width, height or refresh rate differ from the requested values, or if the bit depth is less than 8,
        a warning is issued.

        This function also sets the viewport to the full window size and creates a frame buffer.
        """
        # create a new frame buffer, re-use the old one if one was present, otherwise use a default of range(256)
        # re-use the lookup table if possible, otherwise create a default one ranging from 0 to 255.
        old_lut = self._frame_buffer.lookup_table if self._frame_buffer is not None else range(256)
        self._frame_buffer = FrameBufferPatch(self, old_lut)
        glViewport(0, 0, self._shape[1], self._shape[0])
        # tell openGL to wait for the vertical retrace when swapping buffers (it appears need to do this
        # after creating the frame buffer)
        glfw.swap_interval(1)

        # update the shape property to match the actual value of the window
        (fb_width, fb_height) = glfw.get_framebuffer_size(self._window)
        fb_shape = (fb_height, fb_width)
        if self._shape != fb_shape:
            warnings.warn(f"Actual resolution {fb_shape} does not match requested resolution {self._shape}.")
            self._shape = fb_shape

        (current_size, current_rate, current_bit_depth) = SLM._current_mode(self._monitor_id)

        # verify that the bit depth is at least 8 bit
        if current_bit_depth < 8:
            warnings.warn(
                f"Bit depth is less than 8 bits "
                f"You may not be able to use the full phase resolution of your SLM.")

        # verify the refresh rate is correct
        # Then update the refresh rate to the actual value
        if int(self._refresh_rate) != current_rate:
            warnings.warn(f"Actual refresh rate of {current_rate} Hz does not match set rate "
                          f"of {self._refresh_rate} Hz")
        self._refresh_rate = current_rate

    @staticmethod
    def _init_glfw():
        """Initializes the GLFW library and sets global configuration.

        Note:
            We never de-initialize the library. This should be fine because each slm window releases its resources
            when it is destroyed. If we were to de-initialize the GLFW library (using glfw.terminate()) we run into
            trouble if the user of our library also uses glfw for something else.
        """
        glfw.init()
        glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)  # Required on Mac. Doesn't hurt on Windows
        glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, glfw.TRUE)  # Required on Mac. Useless on Windows
        glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 4)  # request at least opengl 4.2
        glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 2)
        glfw.window_hint(glfw.FLOATING, glfw.TRUE)  # Keep window on top
        glfw.window_hint(glfw.DECORATED, glfw.FALSE)  # Disable window border
        glfw.window_hint(glfw.AUTO_ICONIFY, glfw.FALSE)  # Prevent window minimization during task switch
        glfw.window_hint(glfw.FOCUSED, glfw.FALSE)
        glfw.window_hint(glfw.DOUBLEBUFFER, glfw.TRUE)
        glfw.window_hint(glfw.RED_BITS, 8)  # require at least 8 bits per color channel (256 gray values)
        glfw.window_hint(glfw.GREEN_BITS, 8)
        glfw.window_hint(glfw.BLUE_BITS, 8)
        glfw.window_hint(glfw.COCOA_RETINA_FRAMEBUFFER, glfw.FALSE)  # disable retina multisampling on Mac (untested)
        glfw.window_hint(glfw.SAMPLES, 0)  # disable multisampling

    def _create_window(self):
        """Constructs a new window and associated OpenGL context. Called by SLM.__init__()"""
        # Try to re-use an already existing OpenGL context, so that we can share textures and shaders between
        # SLM objects.
        self._context = Context(self)
        other = next(iter(SLM._active_slms), None)
        shared = other._window if other is not None else None  # noqa: ok to use _window
        SLM._active_slms.add(self)

        monitor = glfw.get_monitors()[self._monitor_id - 1] if self._monitor_id != SLM.WINDOWED else None
        glfw.window_hint(glfw.REFRESH_RATE, int(self._refresh_rate))
        self._window = glfw.create_window(self._shape[1], self._shape[0], "OpenWFS SLM", monitor, shared)
        glfw.set_input_mode(self._window, glfw.CURSOR, glfw.CURSOR_HIDDEN)  # disable cursor
        if monitor:  # full screen mode
            glfw.set_gamma(monitor, 1.0)
        else:  # windowed mode
            glfw.set_window_pos(self._window, self._position[1], self._position[0])

        with self._context:
            self._globals = glGenBuffers(1)  # create buffer for storing globals
            glClearColor(0.0, 0.0, 0.0, 1.0)  # set clear color to black
            self._on_resize()

    @property
    def shape(self) -> tuple[int, int]:
        """Shape (height × width) of the window in pixels.

        Limitations :
            - For windowed-mode SLMs, the size cannot be modified.
            - When moving the SLM to a different monitor (see `monitor_id`), the SLM is sized to match the current
                resolution on that monitor. Note that this value may differ from the value passed as input, because the
                input value is specified in screen coordinates, whereas the reported width is in pixels.
                In this case, the original value of shape will be lost.
            - The `transform` property is not updated automatically, so if the aspect ratio changes
                the transform needs to be set again.
        """
        return self._shape

    @shape.setter
    def shape(self, value: tuple[int, int]):
        if self.monitor_id == SLM.WINDOWED and self._shape != value:
            glfw.set_window_size(self._window, value[1], value[0])
            self._shape = value
            self._on_resize()

    @property
    def position(self) -> tuple[int, int]:
        """
        The position of the top-left corner of the SLM window as (y, x) screen coordinates.

        Note:
            This property is ignored for full-screen SLMs.
        """
        return self._position

    @position.setter
    def position(self, value: tuple[int, int]):
        if self.monitor_id == SLM.WINDOWED and self._position != value:
            glfw.set_window_pos(self._window, value[1], value[0])
            self._position = value

    @property
    def refresh_rate(self) -> Quantity[u.Hz]:
        """
        Refresh rate of the SLM in Hz (read only).

        Note:
            The refresh rate cannot be modified after the SLM is created. When moving
            the SLM to a different monitor (see `monitor_id`), the refresh rate is changed to the current video
            mode on that monitor. Note that OpenGL specifies this value as an integer, whereas some SLMs support
            non-integer refresh rates. It is always best to not specify the refresh rate and set the video mode
            in the operating system before creating the SLM object.
        """
        return self._refresh_rate * u.Hz

    @property
    def period(self) -> Quantity[u.ms]:
        """The period of the refresh rate in milliseconds (read only).
        """
        return (1000 / self._refresh_rate) * u.ms

    @property
    def monitor_id(self) -> int:
        """
        Number of the monitor (1 for primary screen, 2 for secondary screen, etc.) for the SLM.

        Each monitor can only hold a single full-screen window.
        Use monitor_id=SLM.WINDOWED to show a windowed SLM on the primary screen
        (for debugging and monitoring purposes).
        There can be multiple windowed SLMs on the
        primary screen, but there cannot also be a fullscreen SLM on the primary screen at the same time.

        `monitor_id` can be modified at run time, in which case the current SLM window is replaced by a new window
        on a different monitor. When moving the SLM to a different window, width, height and refresh_rate are set
        """
        return self._monitor_id

    @monitor_id.setter
    def monitor_id(self, value):
        if value == self._monitor_id:
            return

        self._assert_window_available(value)
        self._monitor_id = value
        (self._shape, self._refresh_rate, _) = SLM._current_mode(value)
        monitor = glfw.get_monitors()[value - 1] if value != SLM.WINDOWED else None

        # move window to new monitor
        glfw.set_window_monitor(self._window, monitor, self._position[1], self._position[0], self._shape[1],
                                self._shape[0],
                                int(self._refresh_rate))
        self._on_resize()

    def __del__(self):
        """Destructor for the SLM object. This function destroys the window and releases all resources."""
        glfw.destroy_window(self._window)

    def update(self):
        """Sends the new phase pattern to be displayed on the SLM.

        This function waits for the vsync, and returns directly after it.
        Therefore, it can be used as software synchronization to the SLM.

        Note:
            This function *does not* wait for the image to appear on the SLM.
            To wait for the image stabilization explicitly, use 'wait()'.
            However, this should rarely be needed since all Detectors
            already call wait_finished before starting a measurement.

        Note:
            At the moment, :meth:`~.SLM.update` blocks until all OpenGL commands are processed,
            and a vertical retrace occurs (i. e., the hardware signals the start of a new frame).
            This behavior may change in the future and should not be relied on.
            Instead, use the automatic synchronization mechanism to synchronize detectors with
            the SLM hardware.
        """
        with self._context:
            # first draw all patches into the frame buffer
            glBindFramebuffer(GL.GL_FRAMEBUFFER, self._frame_buffer._frame_buffer)  # noqa - ok to access 'friend class'
            glClear(GL.GL_COLOR_BUFFER_BIT)
            for patch in self.patches:
                patch._draw()  # noqa - ok to access 'friend class'

            # then draw the frame buffer to the screen
            glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)
            self._frame_buffer._draw()  # noqa - ok to access 'friend class'

            glfw.poll_events()  # process window messages

            # start 'moving' phase, then display the newly rendered image
            self._start()
            glfw.swap_buffers(self._window)

            # wait for buffer swap to complete (this should be directly after a vsync, so returning from this
            # function _should_ be synced with the vsync)
            glFinish()

        # call _start again to update the _end_time_ns property,
        # since some time has passed waiting for the vsync
        self._start()

    @property
    def latency(self) -> Quantity[u.ms]:
        """Latency (a.k.a. 'idle time')

        Represents the time delay between the vertical retrace
        and the start of the SLM response to then new frame.
        """
        return self._latency

    @Device.latency.setter
    def latency(self, value: Quantity[u.ms]):
        self._latency = value.to(u.ms)

    @Device.duration.setter
    def duration(self, value: Quantity[u.ms]):
        self._duration = value.to(u.ms)

    @property
    def coordinate_system(self) -> str:
        """Specifies the base coordinate system that is used to map vertex coordinates to the SLM window.

            Possible values are 'full', 'short' and 'long'.

            'full' means that the coordinate range (-1,-1) to (1,1) is mapped to the entire SLM window.
            If the window is not square, this means that the coordinates are anisotropic.

            'short' and 'long' map the coordinate range (-1,-1) to (1,1) to a square.
            'short' means that the square is scaled to fill the short side of the SLM (introducing zero-padding at the
            edges).

            'long' means that the square is scaled to fill the long side of the SLM
            (causing part of the coordinate range to be cropped because these coordinates correspond to points outside
            the SLM window).

            For a square SLM, 'full', 'short' and 'long' are all equivalent.

            In all three cases, (-1,-1) corresponds to the top-left corner of the screen, and (1,-1) to the
            bottom-left corner. This convention is consistent with that used in numpy/matplotlib

            To further modify the mapping system, use the `transform` property.
        """
        return self._coordinate_system

    @coordinate_system.setter
    def coordinate_system(self, value: str):
        if value not in ['full', 'short', 'long']:
            raise ValueError(f"Unsupported coordinate system {value}")
        self._coordinate_system = value
        self.transform = self._transform  # trigger update of transform matrix on gpu

    @property
    def transform(self) -> Transform:
        """Global transformation matrix

        The transform determines how the vertex coordinates that make up the shape of a Patch (see
        :class:`Patch`) are mapped to the standard coordinate system.
        In turn, the `coordinate_system` property determines how this coordinate system is mapped to the SLM window.
        By default, this value is just the identity transformation `Transform()`.
        """
        return self._transform

    @transform.setter
    def transform(self, value: Transform):
        # first compute the basic coordinate transform
        width = self._shape[1]
        height = self._shape[0]
        if not isinstance(value, Transform):
            raise ValueError("Transform must be a Transform object")
        self._transform = value
        self._pixel_reader = None
        self._phase_reader = None
        self._field_reader = None

        # update matrix stored on the gpu
        if self._coordinate_system == 'full':
            transform = self._transform
        else:
            scale_width = (width > height) == (self._coordinate_system == 'short')
            if scale_width:
                root_transform = Transform(np.array(((1.0, 0.0), (0.0, height / width))))
            else:
                root_transform = Transform(np.array(((width / height, 0.0), (0.0, 1.0))))
            transform = self._transform @ root_transform

        padded = transform.opencl_matrix()
        with self._context:
            glBindBuffer(GL.GL_UNIFORM_BUFFER, self._globals)
            glBufferData(GL.GL_UNIFORM_BUFFER, padded.size * 4, padded, GL.GL_STATIC_DRAW)
            glBindBufferBase(GL.GL_UNIFORM_BUFFER, 1, self._globals)  # connect buffer to binding point 1

    @property
    def lookup_table(self) -> Sequence[int]:
        """Lookup table that is used to map the wrapped phase range of 0-2pi to gray values
        (represented in a range from 0 to 256). By default, this is just range(256).
        Note that the lookup table need not contain 256 elements.
        A typical scenario is to use something like `slm.lookup_table=range(142)` to map the 0-2pi range
        to only the first 142 gray values of the slm.
        """
        return self._frame_buffer.lookup_table

    @lookup_table.setter
    def lookup_table(self, value: Sequence[int]):
        self._frame_buffer.lookup_table = value[:]

    def set_phases(self, values: ArrayLike, update=True):
        self.primary_patch.set_phases(values, update)

    @property
    def pixels(self) -> Detector:
        """Returns a 'camera' to monitor the current value of the pixels displayed on  the SLM."""
        if self._pixel_reader is None:
            self._pixel_reader = FrontBufferReader(self)
        return self._pixel_reader

    @property
    def field(self) -> Detector:
        """Returns a 'camera' to monitor the current field coming from the SLM.

        Returns:
             a detector that returns `self.amplitude * exp(1.0j * self.phases.read()`
        """
        if self._field_reader is None:
            self._field_reader = PhaseToField(self.phases)
        return self._field_reader

    @property
    def phases(self) -> Detector:
        """Returns an object to monitor the phase pattern last sent to the SLM.

        This is the pattern before applying the lookup table.
        """
        if self._phase_reader is None:
            self._phase_reader = FrameBufferReader(self)
        return self._phase_reader


class FrontBufferReader(Detector):
    def __init__(self, slm):
        self._context = Context(slm)
        super().__init__(data_shape=None, pixel_size=None, duration=0.0 * u.ms, latency=0.0 * u.ms,
                         multi_threaded=False)

    @property
    def data_shape(self):
        return self._context.slm.shape

    def _fetch(self, *args, **kwargs) -> np.ndarray:
        with self._context:
            glReadBuffer(GL.GL_FRONT)
            shape = self.data_shape
            data = np.empty(shape, dtype='uint8')
            glReadPixels(0, 0, shape[1], shape[0], GL.GL_RED, GL.GL_UNSIGNED_BYTE, data)
            # flip data upside down, because the OpenGL convention is to have the origin at the bottom left,
            # but we want it at the top left (like in numpy)
            return data[::-1, :]


class FrameBufferReader(Detector):
    def __init__(self, slm):
        self._context = Context(slm)
        super().__init__(data_shape=None, pixel_size=None, duration=0.0 * u.ms, latency=0.0 * u.ms,
                         multi_threaded=False)

    @property
    def data_shape(self):
        return self._context.slm.shape

    def _fetch(self, *args, **kwargs) -> np.ndarray:
        with self._context as slm:
            return slm._frame_buffer.get_pixels()  # noqa - ok to access 'friend class'



import warnings
from typing import Optional, Union

import astropy.units as u
import numpy as np
from astropy.units import Quantity
from numpy.typing import ArrayLike
from scipy.signal import fftconvolve

from ..core import Processor, Detector
from ..plot_utilities import imshow  # noqa - for debugging
from ..processors import TransformProcessor
from ..simulation.mockdevices import XYStage, Camera, StaticSource
from ..utilities import project, place, Transform, get_pixel_size, patterns, CoordinateType


class Microscope(Processor):
    """A simulated microscope with pupil-conjugate SLM.

    The microscope simulates physical effects such as aberrations and noise, as well as devices typically found in a
    wavefront shaping microscope: a spatial light modulator, translation stages, and a camera.
    This simulation is designed to test algorithms for wavefront shaping and alignment.

    The simulation takes the field at the SLM, applies the phase aberrations, and masks the field with
    a pupil function corresponding to the numerical aperture of the microscope objective.
    This field is then Fourier transformed to obtain the intensity point spread function,
    with which the source image is convolved.

    Finally, the resulting image is mapped to the camera using a magnification factor, or affine transformation matrix.
    The propagation is normalized such that a pupil fully filled with a field strength of 1.0 will produce an image
    that has the same total intensity as the source image.
    """

    def __init__(self, source: Union[Detector, np.ndarray], *, data_shape=None,
                 numerical_aperture: float = 1.0,
                 wavelength: Quantity[u.nm],
                 magnification: float = 1.0, xy_stage=None, z_stage=None,
                 incident_field: Union[Detector, ArrayLike, None] = None,
                 incident_transform: Optional[Transform] = None,
                 aberrations: Union[Detector, np.ndarray, None] = None,
                 aberration_transform: Optional[Transform] = None,
                 multi_threaded: bool = True):
        """
        Args:
            source: 2-D image (must have `pixel_size` metadata), or
                a detector that produces 2-D images of the original 'specimen'
            data_shape: shape (size in pixels) of the output.
                Default value: source.data_shape
            numerical_aperture: Numerical aperture of the microscope objective
            wavelength: Wavelength of the light used for imaging,
                the wavelength and numerical_aperture together determine the resolution of the microscope.
            magnification: Scalar magnification factor between input and output image.
                Note that this factor does not affect the effective image resolution.
                Increasing the magnification will just produce a zoomed-in blurred image.
                The default settings for data_shape, pixel_size and magnification cause the simulated microscope
                to only convolve the source image with the point-spread function (PSF) without applying any scaling.
            xy_stage (XYStage): Optional stage object that can be used to move the sample laterally.
                Defaults to a MockXYStage.
            z_stage (Stage): Optional stage object that moves the sample up and down to focus the microscope.
                Higher values are further away from the microscope objective.
                Defaults to a MockStage.
            incident_field: Produces 2-D complex images containing the field output of the SLM.
                If no `slm_transform` is specified, the `pixel_size` attribute should
                 correspond to normalized pupil coordinates
                (e. g. with a disk of radius 1.0, i. e. an extent of 2.0, corresponding to an NA of 1.0)
            incident_transform (Optional[Transform]):
                Optional Transform that transforms the phase pattern from the slm object
                (in slm.pixel_size units) to normalized pupil coordinates.
                Typically, the slm image is already in normalized pupil coordinates,
                but this transform can be used to mimic SLM misalignment.
            aberrations: 2-D image containing the phase (in radians) of aberrations observed
                in the back pupil of the microscope objective, or a Detector object that automatically produces such
                images. The `extent` attribute corresponds to normalized pupil coordinates. For example, with a
                numerical aperture of 0.6, the extent of the image should be 1.2. If a 2-D image without pixel_size
                metadata is provided, the extent is automatically set to 2.0 * numerical_aperture.
            aberration_transform (Optional[Transform]):
                Optional Transform that transforms the phase pattern from the aberration object
                (in slm.pixel_size units) to normalized pupil coordinates.
                Typically, the slm image is already in normalized pupil coordinates,
                but this transform may e. g., be used to scale an aberration pattern
                from extent 2.0 to 2.0 * NA.

        Note:
            The aberration map and slm phase map are cropped/padded to the NA of the microscope objective, and
            scaled to have the same pixel resolution so that they can be added.
        """
        if not isinstance(source, Detector):
            if get_pixel_size(source) is None:
                raise ValueError("The source must have a pixel_size attribute.")
            source = StaticSource(source)

        if aberrations is not None and not isinstance(aberrations, Detector):
            if get_pixel_size(aberrations) is None:
                aberrations = StaticSource(aberrations)

        super().__init__(source, aberrations, incident_field, multi_threaded=multi_threaded)
        self._magnification = magnification
        self._data_shape = data_shape if data_shape is not None else source.data_shape
        self.numerical_aperture = numerical_aperture
        self.aberration_transform = aberration_transform
        self.slm_transform = incident_transform
        self.wavelength = wavelength.to(u.nm)
        self.oversampling_factor = 2.0
        self.xy_stage = xy_stage or XYStage(0.1 * u.um, 0.1 * u.um)
        self.z_stage = z_stage  # or MockStage()
        self._psf = None

    def _fetch(self, source: np.ndarray, aberrations: np.ndarray,  # noqa
               incident_field: np.ndarray) -> np.ndarray:
        """
        Updates the image on the camera sensor

        To compute the image:
        * First trigger the source, slm, and aberration sources
        * Then read the corresponding images.
        * Combines the slm and aberration images to compute the PSF
        * Crop the source image (not implemented yet) and upsample if needed (not implemented yet)
        * Convolve the source image with the PSF.
        * Compute the magnified and cropped image on the camera.

        Args:
            source:
            aberrations:
            incident_field:

        Returns:

        """

        # First crop and downscale the source image to have the same size as the output
        # todo: add some padding
        # todo: add option for oversampling
        source_pixel_size = get_pixel_size(source)
        target_pixel_size = self.pixel_size / self.magnification
        if np.any(source_pixel_size > target_pixel_size):
            warnings.warn("The resolution of the specimen image is worse than that of the output.")

        # Note: there seems to be a bug (feature?) in `fftconvolve` that shifts the image by one pixel
        # when the 'same' option is used. To compensate for this feature,
        # the image is shifted by `-source_pixel_size` here.
        # TODO: this seems to add an emtpy row and column to the image, which is not what we want.
        shift = Quantity((self.xy_stage.y, self.xy_stage.x)) - source_pixel_size
        source = place(self.data_shape, target_pixel_size, source, shift)

        # Calculate the field in the pupil plane.
        #
        # First, set up pupil coordinates such that:
        # 1. the Fourier transform of the pupil has a resolution
        #    that exactly matches the resolution of the specimen image.
        # 2. the resolution in the pupil plane is high enough such that
        #    the Fourier transform of the pupil field has a size that is at least equal to the fov of the microscope.
        #    This means that then number of pixels should be at least as high as the number of points in the fov
        #    (at the resolution of the specimen image).
        #
        # The NA of the pupil corresponds to a disk that is contained in this pupil plane.
        # We compute the aberrations over the full pupil plane, and clip to the NA by using a disk function.
        # TODO: think about what happens when the requested output resolution is lower than the diffraction limit
        #       at the moment, not the full pupil is used.
        # TODO: think about what happens when the slm is smaller than the pupil

        # condition 1. Extent of pupil in pupil coordinates: Abbe limit should give pixel_size resolution
        pupil_extent = self.wavelength / target_pixel_size

        # condition 2. Minimum number of pixels in x and y should be data_shape
        pupil_shape = self.data_shape

        # Compute the field in the pupil plane
        # The aberrations and the SLM phase pattern are both mapped to the pupil plane coordinates
        pupil_field = patterns.disk(pupil_shape, radius=self.numerical_aperture, extent=pupil_extent)
        pupil_area = np.sum(pupil_field)  # TODO (efficiency): compute area directly from radius

        # Project aberrations
        if aberrations is not None:
            # use default of 2.0 * NA for the extent of the aberration map if no pixel size is provided
            aberration_extent = (2.0 * self.numerical_aperture,) * 2 if get_pixel_size(aberrations) is None else None
            pupil_field = pupil_field * np.exp(1.0j * project(aberrations,
                                                              source_extent=aberration_extent,
                                                              out_extent=pupil_extent,
                                                              out_shape=pupil_shape,
                                                              transform=self.aberration_transform))

        # Project SLM fields
        if incident_field is not None:
            pupil_field = pupil_field * project(incident_field, out_extent=pupil_extent, out_shape=pupil_shape,
                                                transform=self.slm_transform)

        # Compute the point spread function
        # This is done by Fourier transforming the pupil field and taking the absolute value squared
        # Due to condition 1, after the Fourier transform,
        # the pixel size matches that of the source (the specimen image).
        # Note: there is no need to `ifftshift` the pupil field, since we are taking the absolute value anyway
        psf = np.abs(np.fft.ifft2(pupil_field)) ** 2
        psf = np.fft.ifftshift(psf) * (psf.size / pupil_area)
        self._psf = psf  # store psf for later inspection

        return fftconvolve(source, psf, 'same')

    @property
    def magnification(self) -> float:
        """
        Magnification from object plane to image plane.

        Note that, as in a real microscope, the magnification does not affect the effective resolution of the image.
        The resolution is determined by the Abbe diffraction limit λ/2NA.
        """
        return self._magnification

    @magnification.setter
    def magnification(self, value: float):
        self._magnification = value

    @property
    def abbe_limit(self) -> Quantity:
        """Returns the Abbe diffraction limit: λ/(2 NA)"""
        return 0.5 * self.wavelength / self.numerical_aperture

    @property
    def pixel_size(self) -> Quantity:
        """Returns the pixel size in the image plane
        This value is always equal to `source.pixel_size * magnification`"""
        return self._sources[0].pixel_size * self.magnification

    @property
    def data_shape(self):
        """Returns the shape of the image in the image plane"""
        return self._data_shape

    def get_camera(self, *, transform: Optional[Transform] = None,
                   data_shape: Optional[tuple[int, int]] = None,
                   pixel_size: Optional[CoordinateType] = None,
                   **kwargs) -> Detector:
        """
        Returns a simulated camera that observes the microscope image.

        The camera is a MockCamera object that simulates an AD-converter with optional noise.
        shot noise and readout noise (see MockCamera for options).
        In addition to the inputs accepted by the MockCamera constructor (data_shape, analog_max, shot_noise, etc.),
        it is also possible to specify a transform, to mimic the (mis)alignment of the camera.

        Args:
            transform ():
            **kwargs ():

        Returns:

        """
        if transform is None and data_shape is None and pixel_size is None:
            src = self
        else:
            src = TransformProcessor(self, data_shape=data_shape, pixel_size=pixel_size, transform=transform)

        return Camera(src, **kwargs)



from typing import Sequence

import numpy as np
from numpy.typing import ArrayLike

from ...utilities import ExtentType, CoordinateType, unitless


class Geometry:
    """Class that represents the shape of a Patch object that can be drawn on the screen.

    The geometry is defined by a set of vertices
    and a set of indices that define the order in which the vertices are drawn.

    The vertices are stored in a numpy array of shape (N, 4) where N is the number of vertices.
    Each vertex is represented by an array of four values [x, y, tx, ty].
    Here, (x, y) are the coordinates that determine where the vertex is drawn on the screen (after the
    applying the `transform` of the SLM that contains this patch).

    The vertices and indices define a triangle strip (see OpenGL specification) that is drawn on the screen.
    A triangle strip is a sequence of connected triangles, where each triangle shares two vertices with the previous
    triangle. The indices are used to determine the order in which the vertices are connected to form the triangles.
    To start a new triangle strip, insert the special index 0xFFFF into the index array.

    (tx, ty) are the texture coordinates that determine which pixel of the texture
    (e. g. the array passed to `set_phases`) is drawn at each vertex.
    For each triangle, the screen coordinates (x,y) define a triangle on the screen, whereas the texture coordinates
    (tx, ty) define a triangle in the texture.
    OpenGL maps the texture triangle onto the screen triangle, using linear interpolation of the coordinates between
    the vertex points.

    The vertex data is uploaded to the GPU when the Geometry object is assigned to the `geometry` property of a Patch.
    """

    def __init__(self, vertices: ArrayLike, indices: ArrayLike):
        """
        Constructs a new Geometry object.

        Args:
            vertices (np.ndarray): The vertices of the geometry. See class documentation for details.
            indices (np.ndarray): The indices of the geometry. See class documentation for details.
        """
        self._vertices = np.array(vertices, dtype=np.float32, copy=False)
        self._indices = np.array(indices, dtype=np.uint16, copy=False)

        if self._vertices.ndim != 2 or self._vertices.shape[1] != 4:
            raise ValueError("Vertices should be a 2-D array with 4 columns")
        if self._indices.ndim != 1:
            raise ValueError("Indices should be a 1-D array")

    @property
    def vertices(self):
        """The vertices of the geometry."""
        return self._vertices

    @property
    def indices(self):
        """The indices of the geometry."""
        return self._indices

    @staticmethod
    def compute_indices_for_grid(shape):
        """
        Computes indices for a rectangular grid of vertices.

        Assuming the vertices represent the corner points of a rectangular grid,
        this function computes the indices needed to connect these vertices into triangle strips.

        Args:
            shape (Sequence[int]): The shape of the grid (nr, nc).
                nr is the number of grid rows, nc is the number of grid columns.
                Note that the number of vertices should equal (nr + 1) · (nc+1).
        """
        i = 0
        nr = shape[0]
        nc = shape[1] + 1
        index_count = nr * (2 * nc + 1)
        indices = np.zeros(index_count, np.uint16)
        for r in range(nr):
            # emit triangle strip (single row)
            row_start = r * nc
            for c in range(nc):
                indices[i] = row_start + c
                indices[i + 1] = row_start + c + nc
                i += 2
            indices[i] = 0xFFFF
            i += 1
        return indices


def rectangle(extent: ExtentType, center: CoordinateType = (0, 0)) -> Geometry:
    """
    Creates a rectangle geometry with the specified extent.

    Args:
        extent (ExtentType): The extent (height, width) of the rectangle.
            Default value is (2, 2).
            If a single value is specified, the same value is used for both axes.
        center (CoordinateType): The center of the rectangle with respect
            to the origin, specified in (y, x) coordinates.
            Default value is (0, 0).
    """
    center = unitless(center)
    extent = unitless(extent)
    if extent.size == 1:
        extent = np.array((extent, extent))

    left = center[1] - 0.5 * extent[1]
    top = center[0] - 0.5 * extent[0]
    right = center[1] + 0.5 * extent[1]
    bottom = center[0] + 0.5 * extent[0]

    vertices = np.array(([left, top, 0.0, 0.0], [right, top, 1.0, 0.0],
                         [left, bottom, 0.0, 1.0], [right, bottom, 1.0, 1.0]), dtype=np.float32)

    indices = Geometry.compute_indices_for_grid((1, 1))
    return Geometry(vertices, indices)


def circular(radii: Sequence[float], segments_per_ring: Sequence[int], edge_count: int = 256,
             center: CoordinateType = (0, 0)) -> Geometry:
    """Creates a circular geometry with the specified extent.

    This geometry maps a texture to a disk or a ring.
    It can be used for feedback-based wavefront shaping to match the segment size to the illumination profile of the
    SLM. [1]

    When used with a linear texture of shape `(1, np.sum(segments_per_ring)` (see `set_phases`),
    this geometry maps the pixels of the texture to concentric rings, starting with the innermost ring,
    and distributing the phase values counter-clockwise, starting at the positive x-axis.
    Each ring will hold the specified `segments_per_ring`.

    Notes:
         When `radii[0] == 0` , a disk is created.
         When `radii[0] > 0`, the center is left open, creating a ring

    Args:
        radii (Sequence[float]): The radii of the rings in the circular geometry.
            The first value is the inner radius of the inner ring, the last value is the outer radius of the outer ring.
        center (CoordinateType): The center of the circle with respect
            to the origin, specified in (y, x) coordinates.
            Default value is (0, 0).
        segments_per_ring (Sequence[int]): The number of segments for each ring.
            The number of segments is one less than the number of rings.
        edge_count (int): The number of edge points to approximate the full circle.
            The more edges, the closer the geometry will approximate a circle.
            Default value is 256.

    [1]: Mastiani, Bahareh, and Ivo M. Vellekoop. "Noise-tolerant wavefront shaping in a Hadamard basis." Optics
    express 29.11 (2021): 17534-17541.
    """
    ring_count = len(radii) - 1
    if len(segments_per_ring) != ring_count:
        raise ValueError(
            "The length of `radii` and `segments_per_ring` should both equal the number of rings (counting "
            "the inner disk as the first ring).")

    # construct coordinates of points on a circle of radius 1.0
    # the start and end point coincide
    angles = np.linspace(0.0, 2 * np.pi, edge_count + 1)
    x = np.cos(angles)
    y = -np.sin(angles)
    x[0] = x[-1]  # make *exactly* equal
    y[0] = y[-1]

    # construct vertices.
    # each ring is a triangle strip with edge_count + 1 vertices on the inside and
    # edge_count + 1 vertices on the outside.
    vertices = np.zeros((ring_count, 2, edge_count + 1, 4), dtype=np.float32)
    x_inside = radii[0]  # coordinates of the vertices at the center of the ring
    y_inside = radii[0]
    segments_inside = 0
    total_segments = np.sum(segments_per_ring)
    for r in range(ring_count):
        x_outside = x * radii[r + 1]  # coordinates of the vertices at the outside of the ring
        y_outside = y * radii[r + 1]
        segments = segments_inside + segments_per_ring[r]
        vertices[r, 0, :, 0] = x_inside + center[1]
        vertices[r, 0, :, 1] = y_inside + center[0]
        vertices[r, 1, :, 0] = x_outside + center[1]
        vertices[r, 1, :, 1] = y_outside + center[0]
        vertices[r, :, :, 2] = np.linspace(segments_inside, segments, edge_count + 1).reshape(
            (1, -1)) / total_segments  # tx
        x_inside = x_outside
        y_inside = y_outside
        segments_inside = segments

    vertices[:, 0, :, 3] = 0.0  # ty inside
    vertices[:, 1, :, 3] = 1.0  # ty outside

    # construct indices for a single ring, and repeat for all rings with the appropriate offset
    indices = Geometry.compute_indices_for_grid((1, edge_count)).reshape((1, -1))
    indices = indices + np.arange(ring_count).reshape((-1, 1)) * vertices.shape[1] * vertices.shape[2]
    indices[:, -1] = 0xFFFF
    return Geometry(vertices.reshape((-1, 4)), indices.reshape(-1))
